{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "---\n",
    "title: \"Week 02 Lab - Student Notebook\"\n",
    "subtitle: \"2024/25 Winter Term\"\n",
    "author: MY NAME (MY CANDIDATE NUMBER)\n",
    "date: 27 January 2025\n",
    "---\n",
    "\n",
    "# ⚙️ Setup\n",
    "\n",
    "**Import required libraries:**"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "\n",
       "            <div id=\"5929qe\"></div>\n",
       "            <script type=\"text/javascript\" data-lets-plot-script=\"library\">\n",
       "                if(!window.letsPlotCallQueue) {\n",
       "                    window.letsPlotCallQueue = [];\n",
       "                }; \n",
       "                window.letsPlotCall = function(f) {\n",
       "                    window.letsPlotCallQueue.push(f);\n",
       "                };\n",
       "                (function() {\n",
       "                    var script = document.createElement(\"script\");\n",
       "                    script.type = \"text/javascript\";\n",
       "                    script.src = \"https://cdn.jsdelivr.net/gh/JetBrains/lets-plot@v4.5.2/js-package/distr/lets-plot.min.js\";\n",
       "                    script.onload = function() {\n",
       "                        window.letsPlotCall = function(f) {f();};\n",
       "                        window.letsPlotCallQueue.forEach(function(f) {f();});\n",
       "                        window.letsPlotCallQueue = [];\n",
       "                        \n",
       "                    };\n",
       "                    script.onerror = function(event) {\n",
       "                        window.letsPlotCall = function(f) {};    // noop\n",
       "                        window.letsPlotCallQueue = [];\n",
       "                        var div = document.createElement(\"div\");\n",
       "                        div.style.color = 'darkred';\n",
       "                        div.textContent = 'Error loading Lets-Plot JS';\n",
       "                        document.getElementById(\"5929qe\").appendChild(div);\n",
       "                    };\n",
       "                    var e = document.getElementById(\"5929qe\");\n",
       "                    e.appendChild(script);\n",
       "                })()\n",
       "            </script>\n",
       "            "
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "import numpy as np\n",
    "import pandas as pd\n",
    "from lets_plot import *\n",
    "LetsPlot.setup_html()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# 📋 Lab Tasks\n",
    "\n",
    "## 🛠 Part 1: Data manipulation with `pandas` (45 min)\n",
    "\n",
    "### ✨ **`pandas` data frame attributes**\n",
    "\n",
    "A `pandas` data frame has a lot of attributes, several of which we shall explore:\n",
    "\n",
    "-   `shape`: shows the number of rows and columns.\n",
    "-   `index`: prints the name of each row in the data frame.\n",
    "-   `columns`: prints the name of each column in the data frame.\n",
    "\n",
    "### 🔧 **`pandas` data frame methods**\n",
    "\n",
    "`pandas` data frames have a lot of methods! We will focus our attention on the following:\n",
    "\n",
    "-   `head` / `tail`: shows the first / last *n* observations of a data frame.\n",
    "-   `to_frame`: converts a series (`pandas` will automatically convert a data frame into a series if one variable is selected) to a data frame.\n",
    "-   `unique`: shows all unique values for qualitative features selected.\n",
    "-   `value_counts`: counts the number of times a unique value appears in a qualitative feature.\n",
    "-   `query`: keeps rows that conform to one or more logical criteria.\n",
    "-   `reset_index`: a subset of a data frame will keep the index of an old data frame. We use this method to change the index to one integer increments.\n",
    "-   `filter`: keeps columns that are included in a user-supplied list (via the `items` parameter).\n",
    "-   `drop`: drops columns that are included in a user-supplied list.\n",
    "-   `rename`: renames already existing columns based on a user-supplied dictionary.\n",
    "-   `assign`: creates a new variable based on a `lambda` function.\n",
    "-   `apply`: applies a `lambda` function across a set of variables.\n",
    "-   `get_dummies`: transforms qualitative features into a series of dummy features.\n",
    "-   `groupby`: perform grouped calculations within qualitative features.\n",
    "\n",
    "**Import `gapminder`**"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [],
   "source": [
    "gapminder = pd.read_csv(\"data/gapminder.csv\")"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### 1.1: Printing `pandas` data frames\n",
    "\n",
    "Let's print `gapminder`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "          country continent  year  lifeExp       pop   gdpPercap\n",
      "0     Afghanistan      Asia  1952   28.801   8425333  779.445314\n",
      "1     Afghanistan      Asia  1957   30.332   9240934  820.853030\n",
      "2     Afghanistan      Asia  1962   31.997  10267083  853.100710\n",
      "3     Afghanistan      Asia  1967   34.020  11537966  836.197138\n",
      "4     Afghanistan      Asia  1972   36.088  13079460  739.981106\n",
      "...           ...       ...   ...      ...       ...         ...\n",
      "1699     Zimbabwe    Africa  1987   62.351   9216418  706.157306\n",
      "1700     Zimbabwe    Africa  1992   60.377  10704340  693.420786\n",
      "1701     Zimbabwe    Africa  1997   46.809  11404948  792.449960\n",
      "1702     Zimbabwe    Africa  2002   39.989  11926563  672.038623\n",
      "1703     Zimbabwe    Africa  2007   43.487  12311143  469.709298\n",
      "\n",
      "[1704 rows x 6 columns]\n",
      "       country continent  year  lifeExp       pop   gdpPercap\n",
      "0  Afghanistan      Asia  1952   28.801   8425333  779.445314\n",
      "1  Afghanistan      Asia  1957   30.332   9240934  820.853030\n",
      "2  Afghanistan      Asia  1962   31.997  10267083  853.100710\n",
      "3  Afghanistan      Asia  1967   34.020  11537966  836.197138\n",
      "4  Afghanistan      Asia  1972   36.088  13079460  739.981106\n",
      "       country continent  year  lifeExp       pop   gdpPercap\n",
      "1699  Zimbabwe    Africa  1987   62.351   9216418  706.157306\n",
      "1700  Zimbabwe    Africa  1992   60.377  10704340  693.420786\n",
      "1701  Zimbabwe    Africa  1997   46.809  11404948  792.449960\n",
      "1702  Zimbabwe    Africa  2002   39.989  11926563  672.038623\n",
      "1703  Zimbabwe    Africa  2007   43.487  12311143  469.709298\n"
     ]
    }
   ],
   "source": [
    "# Printing the object itself\n",
    "print(gapminder)\n",
    "\n",
    "# Printing the first 5 rows\n",
    "print(gapminder.head(5))\n",
    "\n",
    "# Printing the last 5 rows\n",
    "print(gapminder.tail(5))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "We can see the dimensions of our data by calling the shape attribute of `gapminder`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "(1704, 6)"
      ]
     },
     "execution_count": 6,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "gapminder.shape"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Here, we find that `gapminder` has 1,704 rows and 6 columns.\n",
    "\n",
    "### 1.2: Data frames and series\n",
    "\n",
    "We can see that `gapminder` has a list of countries and continents. We will explore continents further. To select only the continent column, we quote the variable name inside brackets next to `gapminder`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "0         Asia\n",
       "1         Asia\n",
       "2         Asia\n",
       "3         Asia\n",
       "4         Asia\n",
       "         ...  \n",
       "1699    Africa\n",
       "1700    Africa\n",
       "1701    Africa\n",
       "1702    Africa\n",
       "1703    Africa\n",
       "Name: continent, Length: 1704, dtype: object"
      ]
     },
     "execution_count": 7,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "gapminder[\"continent\"]"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "👉 **NOTE:** If you have variables separated by any white space or variables that contain special characters such as `-` or `@`, you can only use the brackets notation to select a column. The . (dot) notation can only be used for column names that are valid Python identifiers (e.g., no spaces, must start with a letter or underscore, and contain only alphanumeric characters or underscores).\n",
    "\n",
    "This changes the data frame to a series. If you want the output to remain a data frame, however, you can use the `to_frame` method."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>continent</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>0</th>\n",
       "      <td>Asia</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>Asia</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>Asia</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>3</th>\n",
       "      <td>Asia</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>Asia</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>...</th>\n",
       "      <td>...</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1699</th>\n",
       "      <td>Africa</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1700</th>\n",
       "      <td>Africa</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1701</th>\n",
       "      <td>Africa</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1702</th>\n",
       "      <td>Africa</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1703</th>\n",
       "      <td>Africa</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "<p>1704 rows × 1 columns</p>\n",
       "</div>"
      ],
      "text/plain": [
       "     continent\n",
       "0         Asia\n",
       "1         Asia\n",
       "2         Asia\n",
       "3         Asia\n",
       "4         Asia\n",
       "...        ...\n",
       "1699    Africa\n",
       "1700    Africa\n",
       "1701    Africa\n",
       "1702    Africa\n",
       "1703    Africa\n",
       "\n",
       "[1704 rows x 1 columns]"
      ]
     },
     "execution_count": 8,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "gapminder[\"continent\"].to_frame()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### 1.3: Finding / counting unique values\n",
    "\n",
    "To find the names of all the continents we can use the `unique` method."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "array(['Asia', 'Europe', 'Africa', 'Americas', 'Oceania'], dtype=object)"
      ]
     },
     "execution_count": 7,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "gapminder[\"continent\"].unique()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "To see how many times each continent appears, we can use the `value_counts` method. We use reset index in order to turn `continent` into its own column."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>continent</th>\n",
       "      <th>count</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>0</th>\n",
       "      <td>Africa</td>\n",
       "      <td>624</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>Asia</td>\n",
       "      <td>396</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>Europe</td>\n",
       "      <td>360</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>3</th>\n",
       "      <td>Americas</td>\n",
       "      <td>300</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>Oceania</td>\n",
       "      <td>24</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "</div>"
      ],
      "text/plain": [
       "  continent  count\n",
       "0    Africa    624\n",
       "1      Asia    396\n",
       "2    Europe    360\n",
       "3  Americas    300\n",
       "4   Oceania     24"
      ]
     },
     "execution_count": 8,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "gapminder[\"continent\"].value_counts().reset_index()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "❓**Question:** Do you see anything odd?\n",
    "\n",
    "### 1.4: Performing grouped calculations\n",
    "\n",
    "Suppose we want to calculate average GDP per capita across time. We can use a combination of the `groupby` and `mean` methods from `Pandas`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>year</th>\n",
       "      <th>gdpPercap</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>0</th>\n",
       "      <td>1952</td>\n",
       "      <td>3725.276046</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>1957</td>\n",
       "      <td>4299.408345</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>1962</td>\n",
       "      <td>4725.812342</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>3</th>\n",
       "      <td>1967</td>\n",
       "      <td>5483.653047</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>1972</td>\n",
       "      <td>6770.082815</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>5</th>\n",
       "      <td>1977</td>\n",
       "      <td>7313.166421</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>6</th>\n",
       "      <td>1982</td>\n",
       "      <td>7518.901673</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>7</th>\n",
       "      <td>1987</td>\n",
       "      <td>7900.920218</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>8</th>\n",
       "      <td>1992</td>\n",
       "      <td>8158.608521</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>9</th>\n",
       "      <td>1997</td>\n",
       "      <td>9090.175363</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>10</th>\n",
       "      <td>2002</td>\n",
       "      <td>9917.848365</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>11</th>\n",
       "      <td>2007</td>\n",
       "      <td>11680.071820</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "</div>"
      ],
      "text/plain": [
       "    year     gdpPercap\n",
       "0   1952   3725.276046\n",
       "1   1957   4299.408345\n",
       "2   1962   4725.812342\n",
       "3   1967   5483.653047\n",
       "4   1972   6770.082815\n",
       "5   1977   7313.166421\n",
       "6   1982   7518.901673\n",
       "7   1987   7900.920218\n",
       "8   1992   8158.608521\n",
       "9   1997   9090.175363\n",
       "10  2002   9917.848365\n",
       "11  2007  11680.071820"
      ]
     },
     "execution_count": 9,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "gapminder.groupby(\"year\", as_index=False)[\"gdpPercap\"].mean()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### 1.5: Subsetting rows\n",
    "\n",
    "Suppose we want to investigate as to why Oceania has only 24 observations (see **Part 1.3**), we can start by using the `query` method, which filters rows by one or more conditions."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "metadata": {},
   "outputs": [],
   "source": [
    "gapminder_oceania = gapminder.query(\"continent == 'Oceania'\")"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "📝**Task:** Find the unique values of country in `gapminder_oceania`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Code here"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "❓**Question:** Do you see the reason now?\n",
    "\n",
    "### 1.6: Subsetting columns\n",
    "\n",
    "Suppose we only want our data frame to include country, year and population. We can use the `filter` method in a Pandas data frame setting the `items` parameter equal to a list of feature names that we want to include."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>country</th>\n",
       "      <th>year</th>\n",
       "      <th>pop</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>0</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1952</td>\n",
       "      <td>8425333</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1957</td>\n",
       "      <td>9240934</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1962</td>\n",
       "      <td>10267083</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>3</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1967</td>\n",
       "      <td>11537966</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1972</td>\n",
       "      <td>13079460</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>...</th>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1699</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>1987</td>\n",
       "      <td>9216418</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1700</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>1992</td>\n",
       "      <td>10704340</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1701</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>1997</td>\n",
       "      <td>11404948</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1702</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>2002</td>\n",
       "      <td>11926563</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1703</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>2007</td>\n",
       "      <td>12311143</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "<p>1704 rows × 3 columns</p>\n",
       "</div>"
      ],
      "text/plain": [
       "          country  year       pop\n",
       "0     Afghanistan  1952   8425333\n",
       "1     Afghanistan  1957   9240934\n",
       "2     Afghanistan  1962  10267083\n",
       "3     Afghanistan  1967  11537966\n",
       "4     Afghanistan  1972  13079460\n",
       "...           ...   ...       ...\n",
       "1699     Zimbabwe  1987   9216418\n",
       "1700     Zimbabwe  1992  10704340\n",
       "1701     Zimbabwe  1997  11404948\n",
       "1702     Zimbabwe  2002  11926563\n",
       "1703     Zimbabwe  2007  12311143\n",
       "\n",
       "[1704 rows x 3 columns]"
      ]
     },
     "execution_count": 12,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "gapminder.filter(items=[\"country\",\"year\",\"pop\"])"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Another option is to use the `drop` method which takes a list of features to drop. Here, we specify `axis=1` which signifies columns, not rows (to specify rows, we set `axis=0`)."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 13,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>country</th>\n",
       "      <th>year</th>\n",
       "      <th>pop</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>0</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1952</td>\n",
       "      <td>8425333</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1957</td>\n",
       "      <td>9240934</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1962</td>\n",
       "      <td>10267083</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>3</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1967</td>\n",
       "      <td>11537966</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1972</td>\n",
       "      <td>13079460</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>...</th>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1699</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>1987</td>\n",
       "      <td>9216418</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1700</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>1992</td>\n",
       "      <td>10704340</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1701</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>1997</td>\n",
       "      <td>11404948</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1702</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>2002</td>\n",
       "      <td>11926563</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1703</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>2007</td>\n",
       "      <td>12311143</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "<p>1704 rows × 3 columns</p>\n",
       "</div>"
      ],
      "text/plain": [
       "          country  year       pop\n",
       "0     Afghanistan  1952   8425333\n",
       "1     Afghanistan  1957   9240934\n",
       "2     Afghanistan  1962  10267083\n",
       "3     Afghanistan  1967  11537966\n",
       "4     Afghanistan  1972  13079460\n",
       "...           ...   ...       ...\n",
       "1699     Zimbabwe  1987   9216418\n",
       "1700     Zimbabwe  1992  10704340\n",
       "1701     Zimbabwe  1997  11404948\n",
       "1702     Zimbabwe  2002  11926563\n",
       "1703     Zimbabwe  2007  12311143\n",
       "\n",
       "[1704 rows x 3 columns]"
      ]
     },
     "execution_count": 13,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "gapminder.drop([\"lifeExp\",\"continent\",\"gdpPercap\"],axis=1)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Yet another way to achieve the same result as the `filter` and `drop` methods we've just shown is the double square bracket subsetting:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>country</th>\n",
       "      <th>year</th>\n",
       "      <th>pop</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>0</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1952</td>\n",
       "      <td>8425333</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1957</td>\n",
       "      <td>9240934</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1962</td>\n",
       "      <td>10267083</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>3</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1967</td>\n",
       "      <td>11537966</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1972</td>\n",
       "      <td>13079460</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>...</th>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1699</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>1987</td>\n",
       "      <td>9216418</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1700</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>1992</td>\n",
       "      <td>10704340</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1701</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>1997</td>\n",
       "      <td>11404948</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1702</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>2002</td>\n",
       "      <td>11926563</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1703</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>2007</td>\n",
       "      <td>12311143</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "<p>1704 rows × 3 columns</p>\n",
       "</div>"
      ],
      "text/plain": [
       "          country  year       pop\n",
       "0     Afghanistan  1952   8425333\n",
       "1     Afghanistan  1957   9240934\n",
       "2     Afghanistan  1962  10267083\n",
       "3     Afghanistan  1967  11537966\n",
       "4     Afghanistan  1972  13079460\n",
       "...           ...   ...       ...\n",
       "1699     Zimbabwe  1987   9216418\n",
       "1700     Zimbabwe  1992  10704340\n",
       "1701     Zimbabwe  1997  11404948\n",
       "1702     Zimbabwe  2002  11926563\n",
       "1703     Zimbabwe  2007  12311143\n",
       "\n",
       "[1704 rows x 3 columns]"
      ]
     },
     "execution_count": 10,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "gapminder[[\"country\",\"year\",\"pop\"]]"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "👉 **NOTE:** If you have used `pandas`, you may have used the `loc` and `iloc` methods on data frames. These functions enable users to select both columns and rows in one function. While, in theory, this sounds great, these methods are computationally inefficient, so we advise that you do not use these methods and, instead, opt for `query` and `filter`.\n",
    "\n",
    "### 1.7: Renaming columns\n",
    "\n",
    "It is good practice to convert variable names to \"snake case\" whereby all characters are lower case and each word in the variable is separated by an underscore. To find the variable names expressed as an index, we call the `columns` attribute."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 14,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "Index(['country', 'continent', 'year', 'lifeExp', 'pop', 'gdpPercap'], dtype='object')"
      ]
     },
     "execution_count": 14,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "gapminder.columns"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "From there, we can amend the variable names using a dictionary where the *key* is the current variable name and the *value* is the variable name you would like it to be. We then use the `rename` method, setting the `columns` parameter equal to the dictionary created."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 15,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "Index(['country', 'continent', 'year', 'life_exp', 'pop', 'gdp_per_cap'], dtype='object')"
      ]
     },
     "execution_count": 15,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "# Create a dictionary of variable names using snake case\n",
    "snake_case_var_names = {\n",
    "  \"country\":\"country\",\n",
    "  \"continent\":\"continent\",\n",
    "  \"year\":\"year\",\n",
    "  \"lifeExp\": \"life_exp\",\n",
    "  \"pop\":\"pop\",\n",
    "  \"gdpPercap\":\"gdp_per_cap\"\n",
    "}\n",
    "\n",
    "# Set the columns attribute to this list\n",
    "gapminder = gapminder.rename(columns=snake_case_var_names)\n",
    "\n",
    "# Check the columns attribute\n",
    "gapminder.columns"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### 1.8: Creating new variables\n",
    "\n",
    "We know that Gross Domestic Product can be obtained from multiplying GDP per capita and population. To do this in Pandas, we simply insert `*` between the `gdp_per_cap` and `pop` columns found in `gapminder`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 16,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>country</th>\n",
       "      <th>continent</th>\n",
       "      <th>year</th>\n",
       "      <th>life_exp</th>\n",
       "      <th>pop</th>\n",
       "      <th>gdp_per_cap</th>\n",
       "      <th>gdp</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>0</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>Asia</td>\n",
       "      <td>1952</td>\n",
       "      <td>28.801</td>\n",
       "      <td>8425333</td>\n",
       "      <td>779.445314</td>\n",
       "      <td>6.567086e+09</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>Asia</td>\n",
       "      <td>1957</td>\n",
       "      <td>30.332</td>\n",
       "      <td>9240934</td>\n",
       "      <td>820.853030</td>\n",
       "      <td>7.585449e+09</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>Asia</td>\n",
       "      <td>1962</td>\n",
       "      <td>31.997</td>\n",
       "      <td>10267083</td>\n",
       "      <td>853.100710</td>\n",
       "      <td>8.758856e+09</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>3</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>Asia</td>\n",
       "      <td>1967</td>\n",
       "      <td>34.020</td>\n",
       "      <td>11537966</td>\n",
       "      <td>836.197138</td>\n",
       "      <td>9.648014e+09</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>Asia</td>\n",
       "      <td>1972</td>\n",
       "      <td>36.088</td>\n",
       "      <td>13079460</td>\n",
       "      <td>739.981106</td>\n",
       "      <td>9.678553e+09</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>...</th>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1699</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>Africa</td>\n",
       "      <td>1987</td>\n",
       "      <td>62.351</td>\n",
       "      <td>9216418</td>\n",
       "      <td>706.157306</td>\n",
       "      <td>6.508241e+09</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1700</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>Africa</td>\n",
       "      <td>1992</td>\n",
       "      <td>60.377</td>\n",
       "      <td>10704340</td>\n",
       "      <td>693.420786</td>\n",
       "      <td>7.422612e+09</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1701</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>Africa</td>\n",
       "      <td>1997</td>\n",
       "      <td>46.809</td>\n",
       "      <td>11404948</td>\n",
       "      <td>792.449960</td>\n",
       "      <td>9.037851e+09</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1702</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>Africa</td>\n",
       "      <td>2002</td>\n",
       "      <td>39.989</td>\n",
       "      <td>11926563</td>\n",
       "      <td>672.038623</td>\n",
       "      <td>8.015111e+09</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1703</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>Africa</td>\n",
       "      <td>2007</td>\n",
       "      <td>43.487</td>\n",
       "      <td>12311143</td>\n",
       "      <td>469.709298</td>\n",
       "      <td>5.782658e+09</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "<p>1704 rows × 7 columns</p>\n",
       "</div>"
      ],
      "text/plain": [
       "          country continent  year  life_exp       pop  gdp_per_cap  \\\n",
       "0     Afghanistan      Asia  1952    28.801   8425333   779.445314   \n",
       "1     Afghanistan      Asia  1957    30.332   9240934   820.853030   \n",
       "2     Afghanistan      Asia  1962    31.997  10267083   853.100710   \n",
       "3     Afghanistan      Asia  1967    34.020  11537966   836.197138   \n",
       "4     Afghanistan      Asia  1972    36.088  13079460   739.981106   \n",
       "...           ...       ...   ...       ...       ...          ...   \n",
       "1699     Zimbabwe    Africa  1987    62.351   9216418   706.157306   \n",
       "1700     Zimbabwe    Africa  1992    60.377  10704340   693.420786   \n",
       "1701     Zimbabwe    Africa  1997    46.809  11404948   792.449960   \n",
       "1702     Zimbabwe    Africa  2002    39.989  11926563   672.038623   \n",
       "1703     Zimbabwe    Africa  2007    43.487  12311143   469.709298   \n",
       "\n",
       "               gdp  \n",
       "0     6.567086e+09  \n",
       "1     7.585449e+09  \n",
       "2     8.758856e+09  \n",
       "3     9.648014e+09  \n",
       "4     9.678553e+09  \n",
       "...            ...  \n",
       "1699  6.508241e+09  \n",
       "1700  7.422612e+09  \n",
       "1701  9.037851e+09  \n",
       "1702  8.015111e+09  \n",
       "1703  5.782658e+09  \n",
       "\n",
       "[1704 rows x 7 columns]"
      ]
     },
     "execution_count": 16,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "gapminder[\"gdp\"] = gapminder[\"gdp_per_cap\"] * gapminder[\"pop\"]\n",
    "gapminder"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "After having calculated GDP, you may be interested in coding whether or not a country has above median GDP. We can turn `where` in Numpy into a function and create a new column using the `assign` method."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 17,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>country</th>\n",
       "      <th>year</th>\n",
       "      <th>gdp</th>\n",
       "      <th>above_median</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>0</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1952</td>\n",
       "      <td>6.567086e+09</td>\n",
       "      <td>False</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1957</td>\n",
       "      <td>7.585449e+09</td>\n",
       "      <td>False</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1962</td>\n",
       "      <td>8.758856e+09</td>\n",
       "      <td>False</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>3</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1967</td>\n",
       "      <td>9.648014e+09</td>\n",
       "      <td>False</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1972</td>\n",
       "      <td>9.678553e+09</td>\n",
       "      <td>False</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>...</th>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1699</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>1987</td>\n",
       "      <td>6.508241e+09</td>\n",
       "      <td>False</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1700</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>1992</td>\n",
       "      <td>7.422612e+09</td>\n",
       "      <td>False</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1701</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>1997</td>\n",
       "      <td>9.037851e+09</td>\n",
       "      <td>False</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1702</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>2002</td>\n",
       "      <td>8.015111e+09</td>\n",
       "      <td>False</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1703</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>2007</td>\n",
       "      <td>5.782658e+09</td>\n",
       "      <td>False</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "<p>1704 rows × 4 columns</p>\n",
       "</div>"
      ],
      "text/plain": [
       "          country  year           gdp  above_median\n",
       "0     Afghanistan  1952  6.567086e+09         False\n",
       "1     Afghanistan  1957  7.585449e+09         False\n",
       "2     Afghanistan  1962  8.758856e+09         False\n",
       "3     Afghanistan  1967  9.648014e+09         False\n",
       "4     Afghanistan  1972  9.678553e+09         False\n",
       "...           ...   ...           ...           ...\n",
       "1699     Zimbabwe  1987  6.508241e+09         False\n",
       "1700     Zimbabwe  1992  7.422612e+09         False\n",
       "1701     Zimbabwe  1997  9.037851e+09         False\n",
       "1702     Zimbabwe  2002  8.015111e+09         False\n",
       "1703     Zimbabwe  2007  5.782658e+09         False\n",
       "\n",
       "[1704 rows x 4 columns]"
      ]
     },
     "execution_count": 17,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "# Returns a boolean array if a quantitative feature is above median values\n",
    "def is_above_median(var):\n",
    "  return np.where(var > np.median(var), True, False)\n",
    "\n",
    "# Apply the function to GDP\n",
    "gapminder[[\"country\",\"year\",\"gdp\"]].assign(above_median = lambda x: is_above_median(x[\"gdp\"]))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "👉 **NOTE:** When using `assign` you can see that we use `lambda x:` followed by the function. All we are doing is using `x` as a placeholder for our data frame (`country`, `year` and `gdp`). In doing so, we can select the column we are interested in using to create our new boolean variable.\n",
    "\n",
    "💁**Tip:** You may have noticed that we can string multiple methods together in Pandas. This is extremely useful, but you might find that your code will get too \"long\". If you find this to be the case, you can use `\\` to spread your code over multiple lines. Here's an example of how to do this with the above code:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 18,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>country</th>\n",
       "      <th>year</th>\n",
       "      <th>gdp</th>\n",
       "      <th>above_median</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>0</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1952</td>\n",
       "      <td>6.567086e+09</td>\n",
       "      <td>False</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1957</td>\n",
       "      <td>7.585449e+09</td>\n",
       "      <td>False</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1962</td>\n",
       "      <td>8.758856e+09</td>\n",
       "      <td>False</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>3</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1967</td>\n",
       "      <td>9.648014e+09</td>\n",
       "      <td>False</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>1972</td>\n",
       "      <td>9.678553e+09</td>\n",
       "      <td>False</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>...</th>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1699</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>1987</td>\n",
       "      <td>6.508241e+09</td>\n",
       "      <td>False</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1700</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>1992</td>\n",
       "      <td>7.422612e+09</td>\n",
       "      <td>False</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1701</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>1997</td>\n",
       "      <td>9.037851e+09</td>\n",
       "      <td>False</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1702</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>2002</td>\n",
       "      <td>8.015111e+09</td>\n",
       "      <td>False</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1703</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>2007</td>\n",
       "      <td>5.782658e+09</td>\n",
       "      <td>False</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "<p>1704 rows × 4 columns</p>\n",
       "</div>"
      ],
      "text/plain": [
       "          country  year           gdp  above_median\n",
       "0     Afghanistan  1952  6.567086e+09         False\n",
       "1     Afghanistan  1957  7.585449e+09         False\n",
       "2     Afghanistan  1962  8.758856e+09         False\n",
       "3     Afghanistan  1967  9.648014e+09         False\n",
       "4     Afghanistan  1972  9.678553e+09         False\n",
       "...           ...   ...           ...           ...\n",
       "1699     Zimbabwe  1987  6.508241e+09         False\n",
       "1700     Zimbabwe  1992  7.422612e+09         False\n",
       "1701     Zimbabwe  1997  9.037851e+09         False\n",
       "1702     Zimbabwe  2002  8.015111e+09         False\n",
       "1703     Zimbabwe  2007  5.782658e+09         False\n",
       "\n",
       "[1704 rows x 4 columns]"
      ]
     },
     "execution_count": 18,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "gapminder[[\"country\",\"year\",\"gdp\"]].\\\n",
    "    assign(above_median = lambda x: is_above_median(x[\"gdp\"]))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "📝**Task:** This output is not very helpful. Try using some of the commands we have gone over to create a more useful data frame."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 19,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>country</th>\n",
       "      <th>year</th>\n",
       "      <th>gdp</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>0</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>2007</td>\n",
       "      <td>3.107929e+10</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>Algeria</td>\n",
       "      <td>1952</td>\n",
       "      <td>2.272563e+10</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>Algeria</td>\n",
       "      <td>1957</td>\n",
       "      <td>3.095611e+10</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>3</th>\n",
       "      <td>Algeria</td>\n",
       "      <td>1962</td>\n",
       "      <td>2.806140e+10</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>Algeria</td>\n",
       "      <td>1967</td>\n",
       "      <td>4.143324e+10</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>...</th>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>847</th>\n",
       "      <td>Vietnam</td>\n",
       "      <td>2007</td>\n",
       "      <td>2.081746e+11</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>848</th>\n",
       "      <td>Yemen, Rep.</td>\n",
       "      <td>1992</td>\n",
       "      <td>2.512511e+10</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>849</th>\n",
       "      <td>Yemen, Rep.</td>\n",
       "      <td>1997</td>\n",
       "      <td>3.351236e+10</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>850</th>\n",
       "      <td>Yemen, Rep.</td>\n",
       "      <td>2002</td>\n",
       "      <td>4.179396e+10</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>851</th>\n",
       "      <td>Yemen, Rep.</td>\n",
       "      <td>2007</td>\n",
       "      <td>5.065987e+10</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "<p>852 rows × 3 columns</p>\n",
       "</div>"
      ],
      "text/plain": [
       "         country  year           gdp\n",
       "0    Afghanistan  2007  3.107929e+10\n",
       "1        Algeria  1952  2.272563e+10\n",
       "2        Algeria  1957  3.095611e+10\n",
       "3        Algeria  1962  2.806140e+10\n",
       "4        Algeria  1967  4.143324e+10\n",
       "..           ...   ...           ...\n",
       "847      Vietnam  2007  2.081746e+11\n",
       "848  Yemen, Rep.  1992  2.512511e+10\n",
       "849  Yemen, Rep.  1997  3.351236e+10\n",
       "850  Yemen, Rep.  2002  4.179396e+10\n",
       "851  Yemen, Rep.  2007  5.065987e+10\n",
       "\n",
       "[852 rows x 3 columns]"
      ]
     },
     "execution_count": 19,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "gapminder[[\"country\",\"year\",\"gdp\"]].\\\n",
    "  assign(above_median = lambda x: is_above_median(x[\"gdp\"])).\\\n",
    "    query(\"above_median == True\").\\\n",
    "      drop(\"above_median\", axis=1).\\\n",
    "        reset_index(drop=True)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### 1.9: Preparing data for `scikit-learn`, an example\n",
    "\n",
    "📝**Task:** Filter the data to only include observations from 2007."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 20,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>country</th>\n",
       "      <th>continent</th>\n",
       "      <th>year</th>\n",
       "      <th>life_exp</th>\n",
       "      <th>pop</th>\n",
       "      <th>gdp_per_cap</th>\n",
       "      <th>gdp</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>0</th>\n",
       "      <td>Afghanistan</td>\n",
       "      <td>Asia</td>\n",
       "      <td>2007</td>\n",
       "      <td>43.828</td>\n",
       "      <td>31889923</td>\n",
       "      <td>974.580338</td>\n",
       "      <td>3.107929e+10</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>Albania</td>\n",
       "      <td>Europe</td>\n",
       "      <td>2007</td>\n",
       "      <td>76.423</td>\n",
       "      <td>3600523</td>\n",
       "      <td>5937.029526</td>\n",
       "      <td>2.137641e+10</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>Algeria</td>\n",
       "      <td>Africa</td>\n",
       "      <td>2007</td>\n",
       "      <td>72.301</td>\n",
       "      <td>33333216</td>\n",
       "      <td>6223.367465</td>\n",
       "      <td>2.074449e+11</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>3</th>\n",
       "      <td>Angola</td>\n",
       "      <td>Africa</td>\n",
       "      <td>2007</td>\n",
       "      <td>42.731</td>\n",
       "      <td>12420476</td>\n",
       "      <td>4797.231267</td>\n",
       "      <td>5.958390e+10</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>Argentina</td>\n",
       "      <td>Americas</td>\n",
       "      <td>2007</td>\n",
       "      <td>75.320</td>\n",
       "      <td>40301927</td>\n",
       "      <td>12779.379640</td>\n",
       "      <td>5.150336e+11</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>...</th>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>137</th>\n",
       "      <td>Vietnam</td>\n",
       "      <td>Asia</td>\n",
       "      <td>2007</td>\n",
       "      <td>74.249</td>\n",
       "      <td>85262356</td>\n",
       "      <td>2441.576404</td>\n",
       "      <td>2.081746e+11</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>138</th>\n",
       "      <td>West Bank and Gaza</td>\n",
       "      <td>Asia</td>\n",
       "      <td>2007</td>\n",
       "      <td>73.422</td>\n",
       "      <td>4018332</td>\n",
       "      <td>3025.349798</td>\n",
       "      <td>1.215686e+10</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>139</th>\n",
       "      <td>Yemen, Rep.</td>\n",
       "      <td>Asia</td>\n",
       "      <td>2007</td>\n",
       "      <td>62.698</td>\n",
       "      <td>22211743</td>\n",
       "      <td>2280.769906</td>\n",
       "      <td>5.065987e+10</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>140</th>\n",
       "      <td>Zambia</td>\n",
       "      <td>Africa</td>\n",
       "      <td>2007</td>\n",
       "      <td>42.384</td>\n",
       "      <td>11746035</td>\n",
       "      <td>1271.211593</td>\n",
       "      <td>1.493170e+10</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>141</th>\n",
       "      <td>Zimbabwe</td>\n",
       "      <td>Africa</td>\n",
       "      <td>2007</td>\n",
       "      <td>43.487</td>\n",
       "      <td>12311143</td>\n",
       "      <td>469.709298</td>\n",
       "      <td>5.782658e+09</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "<p>142 rows × 7 columns</p>\n",
       "</div>"
      ],
      "text/plain": [
       "                country continent  year  life_exp       pop   gdp_per_cap  \\\n",
       "0           Afghanistan      Asia  2007    43.828  31889923    974.580338   \n",
       "1               Albania    Europe  2007    76.423   3600523   5937.029526   \n",
       "2               Algeria    Africa  2007    72.301  33333216   6223.367465   \n",
       "3                Angola    Africa  2007    42.731  12420476   4797.231267   \n",
       "4             Argentina  Americas  2007    75.320  40301927  12779.379640   \n",
       "..                  ...       ...   ...       ...       ...           ...   \n",
       "137             Vietnam      Asia  2007    74.249  85262356   2441.576404   \n",
       "138  West Bank and Gaza      Asia  2007    73.422   4018332   3025.349798   \n",
       "139         Yemen, Rep.      Asia  2007    62.698  22211743   2280.769906   \n",
       "140              Zambia    Africa  2007    42.384  11746035   1271.211593   \n",
       "141            Zimbabwe    Africa  2007    43.487  12311143    469.709298   \n",
       "\n",
       "              gdp  \n",
       "0    3.107929e+10  \n",
       "1    2.137641e+10  \n",
       "2    2.074449e+11  \n",
       "3    5.958390e+10  \n",
       "4    5.150336e+11  \n",
       "..            ...  \n",
       "137  2.081746e+11  \n",
       "138  1.215686e+10  \n",
       "139  5.065987e+10  \n",
       "140  1.493170e+10  \n",
       "141  5.782658e+09  \n",
       "\n",
       "[142 rows x 7 columns]"
      ]
     },
     "execution_count": 20,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "# Code here"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## 📋 Part 2: Data visualisation with `lets_plot`(45 min)\n",
    "\n",
    "They say a picture paints a thousand words, and we are, more often than not, inclined to agree with them (whoever \"they\" are)! Thankfully, we can use perhaps the most widely used package, `lets_plot` (click [here](https://lets-plot.org/) for documentation) to paint such pictures or (more accurately) build such graphs.\n",
    "\n",
    "### 2.1 Histograms\n",
    "\n",
    "Suppose we want to plot the distribution of an outcome of interest. We can use a histogram to plot the distribution of life expectancy in `gapminder`.\n",
    "\n",
    "::: callout-tip\n",
    "**📝 Note: Step-by-step**\n",
    "\n",
    "If you are not familiar with using `lets_plot` for plotting, the line by line construction of the plot below might be useful. The `+` operator is used to add layers to the plot.\n",
    ":::"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "📝**Task:** Create a list of numeric variables (life expectancy, population, GDP per capita, and GDP) and string variables (continent)."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 22,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Code here"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "📝**Task:** Subset the new data frame to only include these variables. **Remember** you can add elements to a list by using `+`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 23,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>life_exp</th>\n",
       "      <th>pop</th>\n",
       "      <th>gdp_per_cap</th>\n",
       "      <th>gdp</th>\n",
       "      <th>continent</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>0</th>\n",
       "      <td>43.828</td>\n",
       "      <td>31889923</td>\n",
       "      <td>974.580338</td>\n",
       "      <td>3.107929e+10</td>\n",
       "      <td>Asia</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>76.423</td>\n",
       "      <td>3600523</td>\n",
       "      <td>5937.029526</td>\n",
       "      <td>2.137641e+10</td>\n",
       "      <td>Europe</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>72.301</td>\n",
       "      <td>33333216</td>\n",
       "      <td>6223.367465</td>\n",
       "      <td>2.074449e+11</td>\n",
       "      <td>Africa</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>3</th>\n",
       "      <td>42.731</td>\n",
       "      <td>12420476</td>\n",
       "      <td>4797.231267</td>\n",
       "      <td>5.958390e+10</td>\n",
       "      <td>Africa</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>75.320</td>\n",
       "      <td>40301927</td>\n",
       "      <td>12779.379640</td>\n",
       "      <td>5.150336e+11</td>\n",
       "      <td>Americas</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>...</th>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>137</th>\n",
       "      <td>74.249</td>\n",
       "      <td>85262356</td>\n",
       "      <td>2441.576404</td>\n",
       "      <td>2.081746e+11</td>\n",
       "      <td>Asia</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>138</th>\n",
       "      <td>73.422</td>\n",
       "      <td>4018332</td>\n",
       "      <td>3025.349798</td>\n",
       "      <td>1.215686e+10</td>\n",
       "      <td>Asia</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>139</th>\n",
       "      <td>62.698</td>\n",
       "      <td>22211743</td>\n",
       "      <td>2280.769906</td>\n",
       "      <td>5.065987e+10</td>\n",
       "      <td>Asia</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>140</th>\n",
       "      <td>42.384</td>\n",
       "      <td>11746035</td>\n",
       "      <td>1271.211593</td>\n",
       "      <td>1.493170e+10</td>\n",
       "      <td>Africa</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>141</th>\n",
       "      <td>43.487</td>\n",
       "      <td>12311143</td>\n",
       "      <td>469.709298</td>\n",
       "      <td>5.782658e+09</td>\n",
       "      <td>Africa</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "<p>142 rows × 5 columns</p>\n",
       "</div>"
      ],
      "text/plain": [
       "     life_exp       pop   gdp_per_cap           gdp continent\n",
       "0      43.828  31889923    974.580338  3.107929e+10      Asia\n",
       "1      76.423   3600523   5937.029526  2.137641e+10    Europe\n",
       "2      72.301  33333216   6223.367465  2.074449e+11    Africa\n",
       "3      42.731  12420476   4797.231267  5.958390e+10    Africa\n",
       "4      75.320  40301927  12779.379640  5.150336e+11  Americas\n",
       "..        ...       ...           ...           ...       ...\n",
       "137    74.249  85262356   2441.576404  2.081746e+11      Asia\n",
       "138    73.422   4018332   3025.349798  1.215686e+10      Asia\n",
       "139    62.698  22211743   2280.769906  5.065987e+10      Asia\n",
       "140    42.384  11746035   1271.211593  1.493170e+10    Africa\n",
       "141    43.487  12311143    469.709298  5.782658e+09    Africa\n",
       "\n",
       "[142 rows x 5 columns]"
      ]
     },
     "execution_count": 23,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "# Code here"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "We will now apply `normalise` to numeric variables"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 24,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Create a function that normalises variables\n",
    "def normalise(var):\n",
    "  return (var-var.mean())/var.std()\n",
    "\n",
    "# Apply normalise function over all numeric variables\n",
    "gapminder_07_normal = gapminder_07.filter(items=num_vars).\\\n",
    "  apply(lambda x: normalise(x), axis=0)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "👉 **NOTE:** Normalising continuous features is good practice in machine learning, and becomes essential when dealing with algorithms that employ some kind of distance measure, such as principle components analysis.\n",
    "\n",
    "Along with normalising features, we need to transform categorical features into one-hot encoded dummy (that is, 0 or 1) features. One-hot simply means that a reference category in a feature will not appear in the transformed data frame. To apply one-hot encoding, we can use `get_dummies` in `Pandas`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 26,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>Americas</th>\n",
       "      <th>Asia</th>\n",
       "      <th>Europe</th>\n",
       "      <th>Oceania</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>0</th>\n",
       "      <td>0</td>\n",
       "      <td>1</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "      <td>1</td>\n",
       "      <td>0</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>3</th>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>1</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>...</th>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>137</th>\n",
       "      <td>0</td>\n",
       "      <td>1</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>138</th>\n",
       "      <td>0</td>\n",
       "      <td>1</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>139</th>\n",
       "      <td>0</td>\n",
       "      <td>1</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>140</th>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>141</th>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "<p>142 rows × 4 columns</p>\n",
       "</div>"
      ],
      "text/plain": [
       "     Americas  Asia  Europe  Oceania\n",
       "0           0     1       0        0\n",
       "1           0     0       1        0\n",
       "2           0     0       0        0\n",
       "3           0     0       0        0\n",
       "4           1     0       0        0\n",
       "..        ...   ...     ...      ...\n",
       "137         0     1       0        0\n",
       "138         0     1       0        0\n",
       "139         0     1       0        0\n",
       "140         0     0       0        0\n",
       "141         0     0       0        0\n",
       "\n",
       "[142 rows x 4 columns]"
      ]
     },
     "execution_count": 26,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "gapminder_07_dummies = pd.get_dummies(gapminder_07[\"continent\"], \n",
    "                                      columns=[\"continent\"], \n",
    "                                      drop_first=True,\n",
    "                                      dtype=int)\n",
    "gapminder_07_dummies"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "After having transformed our continuous and categorical features, we can concatenate the two into a new data frame."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 27,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>life_exp</th>\n",
       "      <th>pop</th>\n",
       "      <th>gdp_per_cap</th>\n",
       "      <th>gdp</th>\n",
       "      <th>Americas</th>\n",
       "      <th>Asia</th>\n",
       "      <th>Europe</th>\n",
       "      <th>Oceania</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>0</th>\n",
       "      <td>-1.919936</td>\n",
       "      <td>-0.082178</td>\n",
       "      <td>-0.832468</td>\n",
       "      <td>-0.288250</td>\n",
       "      <td>0</td>\n",
       "      <td>1</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>0.779886</td>\n",
       "      <td>-0.273813</td>\n",
       "      <td>-0.446584</td>\n",
       "      <td>-0.295646</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "      <td>1</td>\n",
       "      <td>0</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>0.438463</td>\n",
       "      <td>-0.072401</td>\n",
       "      <td>-0.424318</td>\n",
       "      <td>-0.153810</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>3</th>\n",
       "      <td>-2.010799</td>\n",
       "      <td>-0.214066</td>\n",
       "      <td>-0.535216</td>\n",
       "      <td>-0.266522</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>0.688525</td>\n",
       "      <td>-0.025195</td>\n",
       "      <td>0.085483</td>\n",
       "      <td>0.080659</td>\n",
       "      <td>1</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>...</th>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "      <td>...</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>137</th>\n",
       "      <td>0.599815</td>\n",
       "      <td>0.279371</td>\n",
       "      <td>-0.718394</td>\n",
       "      <td>-0.153254</td>\n",
       "      <td>0</td>\n",
       "      <td>1</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>138</th>\n",
       "      <td>0.531315</td>\n",
       "      <td>-0.270983</td>\n",
       "      <td>-0.672999</td>\n",
       "      <td>-0.302674</td>\n",
       "      <td>0</td>\n",
       "      <td>1</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>139</th>\n",
       "      <td>-0.356947</td>\n",
       "      <td>-0.147739</td>\n",
       "      <td>-0.730898</td>\n",
       "      <td>-0.273324</td>\n",
       "      <td>0</td>\n",
       "      <td>1</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>140</th>\n",
       "      <td>-2.039541</td>\n",
       "      <td>-0.218635</td>\n",
       "      <td>-0.809402</td>\n",
       "      <td>-0.300559</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>141</th>\n",
       "      <td>-1.948180</td>\n",
       "      <td>-0.214807</td>\n",
       "      <td>-0.871728</td>\n",
       "      <td>-0.307533</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "      <td>0</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "<p>142 rows × 8 columns</p>\n",
       "</div>"
      ],
      "text/plain": [
       "     life_exp       pop  gdp_per_cap       gdp  Americas  Asia  Europe  \\\n",
       "0   -1.919936 -0.082178    -0.832468 -0.288250         0     1       0   \n",
       "1    0.779886 -0.273813    -0.446584 -0.295646         0     0       1   \n",
       "2    0.438463 -0.072401    -0.424318 -0.153810         0     0       0   \n",
       "3   -2.010799 -0.214066    -0.535216 -0.266522         0     0       0   \n",
       "4    0.688525 -0.025195     0.085483  0.080659         1     0       0   \n",
       "..        ...       ...          ...       ...       ...   ...     ...   \n",
       "137  0.599815  0.279371    -0.718394 -0.153254         0     1       0   \n",
       "138  0.531315 -0.270983    -0.672999 -0.302674         0     1       0   \n",
       "139 -0.356947 -0.147739    -0.730898 -0.273324         0     1       0   \n",
       "140 -2.039541 -0.218635    -0.809402 -0.300559         0     0       0   \n",
       "141 -1.948180 -0.214807    -0.871728 -0.307533         0     0       0   \n",
       "\n",
       "     Oceania  \n",
       "0          0  \n",
       "1          0  \n",
       "2          0  \n",
       "3          0  \n",
       "4          0  \n",
       "..       ...  \n",
       "137        0  \n",
       "138        0  \n",
       "139        0  \n",
       "140        0  \n",
       "141        0  \n",
       "\n",
       "[142 rows x 8 columns]"
      ]
     },
     "execution_count": 27,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "gapminder_07_cleaned = pd.concat([gapminder_07_normal, gapminder_07_dummies], axis=1)\n",
    "gapminder_07_cleaned"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "👉 **NOTE:** These kinds of transformations and concatenations will be employed **a lot** during this course, so please be sure to get used to them.\n",
    "\n",
    "As a final (optional) step, we can convert our Pandas data frame to a Numpy array by employing the `to_numpy` method."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "gapminder_07_cleaned.to_numpy()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "🏆**Challenge:** Which countries had above average life expectancy in 1952?\n",
    "\n",
    "Try these steps: 1. Define a function that returns a boolean if a value in a feature exceeds the average value. 2. Include only the year 1952 by using `query`. 3. Subset the data frame to only include country and life expectancy using `filter`. 4. Create a new boolean variable using the user defined function using `assign`. 5. Include all rows with above average life expectancy using `query`. 6. Subset the data frame to only include country using square bracket indexing. 7. Pull the unique values into an array using `unique`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Code here"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## 📋 Part 2: Data visualisation with `lets_plot`(45 min)\n",
    "\n",
    "They say a picture paints a thousand words, and we are, more often than not, inclined to agree with them (whoever \"they\" are)! Thankfully, we can use perhaps the most widely used package, `lets_plot` (click [here](https://lets-plot.org/) for documentation) to paint such pictures or (more accurately) build such graphs.\n",
    "\n",
    "### 2.1 Histograms\n",
    "\n",
    "Suppose we want to plot the distribution of an outcome of interest. We can use a histogram to plot the distribution of life expectancy in `gapminder`.\n",
    "\n",
    "::: callout-tip\n",
    "**📝 Note: Step-by-step**\n",
    "\n",
    "If you are not familiar with using `lets_plot` for plotting, the line by line construction of the plot below might be useful. The `+` operator is used to add layers to the plot.\n",
    ":::"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "(\n",
    "  ggplot(gapminder, aes(\"life_exp\")) +\n",
    "  geom_histogram()\n",
    ")"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "👥 **WORK IN PAIRS/GROUPS:**\n",
    "\n",
    "Change a few parameters in the `geom_histogram` layer. \n",
    "-   Set `bins` to 40. \n",
    "-   Set `alpha` to 0.75. \n",
    "\n",
    "Afterwards, \n",
    "\n",
    "-   Specify `panel_grid_major_x=element_blank()` in the `theme` function. \n",
    "-   Edit the x and y axes using `labs`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Code here"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### 2.2 Bar graphs\n",
    "\n",
    "We counted the number of countries in each continent that are in `gapminder`. We will do something similar, using a subset of the data in 1997. We can use `lets_plot` to create a bar graph. We first need to specify what we want on the x and y axes, which constitute the first and second argument of the `aes` function."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "gapminder_97 = gapminder.query(\"year == 1997\")\n",
    "\n",
    "(\n",
    "  ggplot(gapminder_97,aes(\"continent\")) +\n",
    "  geom_bar()\n",
    ")"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "👥 **WORK IN PAIRS/GROUPS:**\n",
    "\n",
    "-   Create a new data frame that counts the number of observations for each continent.\n",
    "-   Set `stat=\"identity\"` in `geom_bar`.\n",
    "-   Specify `panel_grid_major_x=element_blank()` in the `theme` function.\n",
    "-   Edit the x and y axes using `labs`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Code here"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "👉 **NOTE:** Don't like vertical bars? Well, we can also create horizontal bars too by specifying `orientation=\"y\"` in `geom_bar`. Remember to switch \"count\" and \"continent\".\n",
    "\n",
    "### 2.3 Scatter plots\n",
    "\n",
    "Scatter plots are the best way to graphically summarise the relationship between two quantitative features. Suppose we wanted to visualise the \"effect\" of GDP per capita on life expectancy. We can use a scatter plot to achieve this."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "(\n",
    "  ggplot(gapminder, aes(\"gdp_per_cap\",\"life_exp\")) +\n",
    "  geom_point()\n",
    ")"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "👥 **WORK IN PAIRS/GROUPS:**\n",
    "\n",
    "-   Create a new feature that takes the natural log of GDP per capita.\n",
    "-   Specify `alpha=0.5` in `geom_point`.\n",
    "-   Edit the x and y axes using `labs`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "gapminder[\"log_gdp_per_cap\"] = np.log(gapminder[\"gdp_per_cap\"])\n",
    "\n",
    "(\n",
    "  ggplot(gapminder, aes(\"log_gdp_per_cap\",\"life_exp\")) +\n",
    "  geom_point(alpha=0.5) +\n",
    "  labs(x=\"Logged GDP per capita (USD)\", y = \"Average life expectancy (years)\")\n",
    ")"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### 2.4 Line plots\n",
    "\n",
    "Line plots can be used to track parameters of interest over time. We will do this by calculating then plotting average GDP per capita for each year."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "gdp_per_cap_by_year = gapminder.groupby(\"year\", as_index=False)[\"gdp_per_cap\"].mean()\n",
    "\n",
    "(\n",
    "  ggplot(gdp_per_cap_by_year, aes(\"year\", \"gdp_per_cap\")) +\n",
    "  geom_line() +\n",
    "  labs(x = \"Year\")\n",
    ")"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "👥 **WORK IN PAIRS/GROUPS:**\n",
    "\n",
    "-   Change `linetype=\"dotted\"` in `geom_line`.\n",
    "-   Add a `geom_point` and `geom_area` layer.\n",
    "-   Change the `breaks` parameter in `scale_x_continuous` to an array from 1952 to 2007 by increments of 10.\n",
    "-   Specify `panel_grid_major_x=element_blank()` in the `theme` function.\n",
    "-   Edit the x and y axes using `labs`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Code here"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "👉 **NOTE:** We think this line plot works in this context, and that it is useful to introduce you to this style of charting progress over time. However, for others contexts filling the area under the line may not work, so do bear this in mind when making decisions!\n",
    "\n",
    "### 2.5 Box plots\n",
    "\n",
    "Box plots are an excellent way to summarise the distribution of values across different strata over key quantities of interest (e.g. median, interquartile range). Suppose we want to show the distribution of populations by continent in `gapminder` in 1982 (excluding Oceania). We can use box plots to graphically illustrate this."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "gapminder_oceania_82 = gapminder.query(\"continent != 'Oceania' & year == 1982\")\n",
    "\n",
    "(\n",
    "  ggplot(gapminder_oceania_82, aes(\"continent\", \"pop\")) +\n",
    "  geom_boxplot() \n",
    ")"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "👥 **WORK IN PAIRS/GROUPS:**\n",
    "\n",
    "-   Add `scale_y_log10` to the plot.\n",
    "-   Specify `panel_grid_major_x=element_blank()` in the `theme` function.\n",
    "-   Edit the x and y axes using `labs`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Code here"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### 2.6 Density plots\n",
    "\n",
    "Finally, we can understand the distribution of continuous features using density plots. Here's an example that looks at the distribution of GDP in 1952."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 28,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "   <div id=\"B340Ly\"></div>\n",
       "   <script type=\"text/javascript\" data-lets-plot-script=\"plot\">\n",
       "   \n",
       "   (function() {\n",
       "   // ----------\n",
       "   \n",
       "   var containerDiv = document.getElementById(\"B340Ly\");\n",
       "   var observer = new ResizeObserver(function(entries) {\n",
       "       for (let entry of entries) {\n",
       "           var width = containerDiv.clientWidth\n",
       "           if (entry.contentBoxSize && width > 0) {\n",
       "           \n",
       "               // Render plot\n",
       "               if (observer) {\n",
       "                   observer.disconnect();\n",
       "                   observer = null;\n",
       "               }\n",
       "\n",
       "               var plotSpec={\n",
       "\"data\":{\n",
       "},\n",
       "\"mapping\":{\n",
       "\"x\":\"gdp\"\n",
       "},\n",
       "\"data_meta\":{\n",
       "\"series_annotations\":[{\n",
       "\"type\":\"str\",\n",
       "\"column\":\"country\"\n",
       "},{\n",
       "\"type\":\"str\",\n",
       "\"column\":\"continent\"\n",
       "},{\n",
       "\"type\":\"int\",\n",
       "\"column\":\"year\"\n",
       "},{\n",
       "\"type\":\"float\",\n",
       "\"column\":\"life_exp\"\n",
       "},{\n",
       "\"type\":\"int\",\n",
       "\"column\":\"pop\"\n",
       "},{\n",
       "\"type\":\"float\",\n",
       "\"column\":\"gdp_per_cap\"\n",
       "},{\n",
       "\"type\":\"float\",\n",
       "\"column\":\"gdp\"\n",
       "}]\n",
       "},\n",
       "\"kind\":\"plot\",\n",
       "\"scales\":[],\n",
       "\"layers\":[{\n",
       "\"geom\":\"density\",\n",
       "\"mapping\":{\n",
       "},\n",
       "\"data_meta\":{\n",
       "},\n",
       "\"data\":{\n",
       "\"..density..\":[2.962780689658235E-11,2.962780689658235E-11,3.3129295847823014E-11,2.7439260016162712E-11,2.7439260016162712E-11,1.8766284685578627E-11,1.2692171638292632E-11,9.873842290999855E-12,8.519594421807653E-12,7.12884331106434E-12,7.12884331106434E-12,5.471836042778219E-12,4.088747169903596E-12,3.2377050713238283E-12,2.730349832900464E-12,2.406807500848717E-12,2.2618784633789097E-12,2.25348121377089E-12,2.1964747103938274E-12,1.918961208828925E-12,1.4318610639579417E-12,9.517575152836015E-13,7.384532667559065E-13,8.00585214732735E-13,9.34701848652265E-13,1.0891805858886541E-12,1.3826509464082436E-12,1.693940227475907E-12,1.6770383498434599E-12,1.3072023406965705E-12,8.917406915922807E-13,5.793785658504116E-13,3.267048675991807E-13,1.3739932809927555E-13,3.9553625998208626E-14,7.553145578274223E-15,1.1532815777366569E-15,2.2830596141469725E-15,1.4931729964549825E-14,6.573167598812587E-14,1.8838319490180586E-13,3.513962491630622E-13,4.2661779726804685E-13,3.371141421207666E-13,1.7350871284767395E-13,5.955291451901203E-14,2.371002675400144E-14,5.4272580463602265E-14,1.6217071812841675E-13,3.2585802660024123E-13,4.2970433118728155E-13,3.873023206184061E-13,2.979109884700152E-13,3.1936864709966507E-13,4.4058116143737396E-13,5.307027142385065E-13,5.517160002895069E-13,5.337556747681014E-13,4.40157184438744E-13,2.6730743062090317E-13,1.1027460387278921E-13,3.005565284742673E-14,6.028620242259094E-15,6.485629092853702E-15,3.2300376919164224E-14,1.1539977045840086E-13,2.688062742027434E-13,4.0853811352478193E-13,4.115952243145694E-13,3.061763384936119E-13,2.60799054669169E-13,3.449418330717935E-13,4.2887587955975803E-13,3.707610260021794E-13,2.1053261688857017E-13,7.790390020673396E-14,1.876529001691663E-14,2.942030298257274E-15,3.002111401510876E-16,1.9938533741057925E-17,8.618788893601428E-19,2.4248554959448603E-20,4.440297550126792E-22,5.292061415517394E-24,4.105104986479402E-26,2.0725749862363816E-28,6.810560060804879E-31,1.4566056596108908E-33,2.0276255255893216E-36,1.8370462809237057E-39,1.0832757879921566E-42,4.15761782923228E-46,1.0385721845289067E-49,1.6885557291802827E-53,1.7868196966942975E-57,2.522369613092246E-60,3.0557082160621493E-56,2.532939784115136E-52,1.3665465836828213E-48,4.7985564628019676E-45,1.0966887145840917E-41,1.6313334611626018E-38,1.5793876589185315E-35,9.952253165706354E-33,4.081698321207794E-30,1.089550941708551E-27,1.892963224877852E-25,2.1405577599692543E-23,1.5754595287650491E-21,7.547387517896586E-20,2.353581956454187E-18,4.778497090321965E-17,6.31973705032492E-16,5.4512345478731245E-15,3.07630570676926E-14,1.144474377966527E-13,2.857314786244439E-13,4.970498837468585E-13,6.414062054814502E-13,6.535031768853861E-13,5.290559570170977E-13,3.2156837452440077E-13,1.3741881994217917E-13,3.959439682781802E-14,7.536499706144642E-15,9.393363216159864E-16,7.638586977725784E-17,4.046797984350978E-18,1.3959298381501364E-19,3.134502133646418E-21,4.581276176849706E-23,4.358142457551549E-25,2.6984048689336576E-27,1.0874293228645292E-29,2.852214691034707E-32,4.869117285844803E-35,5.410095771084824E-38,3.912427496097629E-41,1.841510751846901E-44,5.641424921643583E-48,1.1248379498361383E-51,1.4597473636975618E-55,1.2329686392596978E-59,6.778177731386653E-64,2.425272069820844E-68,5.647998456072378E-73,8.560818002292304E-78,8.44544523953104E-83,5.42271059611632E-88,2.266193007649191E-93,6.164018792790123E-99,1.09123328395129E-104,1.2573548024646622E-110,9.429413410766811E-117,4.602545596256436E-123,1.4621708229556683E-129,3.023324185658473E-136,4.06872103399332E-143,3.5638359207693143E-150,2.0317210426964774E-157,7.5387101779165E-165,1.8206086722924337E-172,2.8616911329367464E-180,2.927625353443648E-188,1.949372486822909E-196,8.448133261850037E-205,2.3829408949375657E-213,4.374741299709023E-222,5.227307791236047E-231,4.0652790349494224E-240,2.057734025473219E-249,6.779141647618266E-259,1.4536062076183658E-268,2.0286420364823755E-278,1.8426831432685312E-288,1.0893877727174308E-298,4.19180354788799E-309,1.0498E-319,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,4.35736147E-316,1.5039423861938623E-305,3.3785114871178175E-295,4.939763893143688E-285,4.7008194800649144E-275,2.911572087494631E-265,1.173729747535062E-255,3.079606779913379E-246,5.259070750860679E-237,5.84533425862254E-228,4.2285968231409557E-219,1.9909932997694795E-210,6.101405775641573E-202,1.2169608585487862E-193,1.579828347923344E-185,1.3348417622436147E-177,7.340679450339782E-170,2.627418933241873E-162,6.120811167452134E-155,9.280582345103682E-148,9.158578383907948E-141,5.882575310333444E-134,2.459198107694482E-127,6.691232600699291E-121,1.1849644714835747E-114,1.365812623218982E-108,1.0246215984795222E-102,5.002907594462341E-97,1.5898935497743046E-91,3.288517856615245E-86,4.427096130534062E-81,3.879040298208283E-76,2.2121584012682063E-71,8.210975235229685E-67,1.9836262836683092E-62,3.118972582590551E-58,3.1919042203949835E-54,2.126056177495887E-50,9.216928356057562E-47,2.6006641951398866E-43,4.776050787494429E-40,5.708738965139674E-37,4.441176415988801E-34,2.248756624235723E-31,7.41094307566829E-29,1.58961184454061E-26,2.2191941241986523E-24,2.0164436429370983E-22,1.192513966722955E-20,4.590156055264834E-19,1.1499475134268658E-17,1.8750600338320317E-16,1.989935165040189E-15,1.3745143152486056E-14,6.179398439309815E-14,1.8081302891206921E-13,3.443497998672145E-13,4.2683157540384974E-13],\n",
       "\"..quantile..\":[0.25,0.5,0.5,0.5,0.75,0.75,0.75,0.75,0.75,0.75,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0],\n",
       "\"gdp\":[5.27846905494405E7,5.27846905494405E7,4.366267348983276E9,8.679750007417112E9,8.679750007417112E9,1.2993232665850948E10,1.7306715324284786E10,2.1620197982718624E10,2.593368064115246E10,3.0247163299586292E10,3.0247163299586292E10,3.456064595802013E10,3.8874128616453964E10,4.31876112748878E10,4.750109393332163E10,5.181457659175547E10,5.612805925018931E10,6.044154190862314E10,6.475502456705698E10,6.906850722549081E10,7.338198988392465E10,7.769547254235849E10,8.200895520079233E10,8.632243785922617E10,9.063592051765999E10,9.494940317609383E10,9.926288583452766E10,1.035763684929615E11,1.0788985115139534E11,1.1220333380982918E11,1.1651681646826302E11,1.2083029912669684E11,1.2514378178513068E11,1.2945726444356451E11,1.3377074710199835E11,1.380842297604322E11,1.4239771241886603E11,1.4671119507729987E11,1.510246777357337E11,1.5533816039416754E11,1.5965164305260138E11,1.6396512571103522E11,1.6827860836946906E11,1.725920910279029E11,1.769055736863367E11,1.8121905634477054E11,1.8553253900320438E11,1.898460216616382E11,1.9415950432007205E11,1.984729869785059E11,2.0278646963693973E11,2.0709995229537357E11,2.114134349538074E11,2.1572691761224124E11,2.2004040027067508E11,2.2435388292910892E11,2.2866736558754276E11,2.329808482459766E11,2.3729433090441043E11,2.4160781356284424E11,2.4592129622127808E11,2.502347788797119E11,2.5454826153814575E11,2.588617441965796E11,2.6317522685501343E11,2.6748870951344727E11,2.718021921718811E11,2.7611567483031494E11,2.804291574887488E11,2.847426401471826E11,2.8905612280561646E11,2.933696054640503E11,2.9768308812248413E11,3.0199657078091797E11,3.063100534393518E11,3.1062353609778564E11,3.149370187562195E11,3.192505014146533E11,3.2356398407308716E11,3.27877466731521E11,3.3219094938995483E11,3.365044320483887E11,3.408179147068225E11,3.4513139736525635E11,3.494448800236901E11,3.5375836268212396E11,3.580718453405578E11,3.6238532799899164E11,3.666988106574255E11,3.710122933158593E11,3.7532577597429315E11,3.79639258632727E11,3.839527412911608E11,3.8826622394959467E11,3.925797066080285E11,3.9689318926646234E11,4.012066719248962E11,4.0552015458333E11,4.0983363724176385E11,4.141471199001977E11,4.1846060255863153E11,4.2277408521706537E11,4.270875678754992E11,4.3140105053393304E11,4.357145331923669E11,4.400280158508007E11,4.4434149850923456E11,4.486549811676684E11,4.5296846382610223E11,4.572819464845361E11,4.615954291429699E11,4.6590891180140375E11,4.702223944598376E11,4.745358771182714E11,4.7884935977670526E11,4.8316284243513904E11,4.874763250935729E11,4.917898077520067E11,4.9610329041044055E11,5.004167730688744E11,5.047302557273082E11,5.0904373838574207E11,5.133572210441759E11,5.1767070370260974E11,5.219841863610436E11,5.262976690194774E11,5.3061115167791125E11,5.349246343363451E11,5.3923811699477893E11,5.4355159965321277E11,5.478650823116466E11,5.521785649700804E11,5.564920476285143E11,5.608055302869481E11,5.65119012945382E11,5.694324956038158E11,5.737459782622496E11,5.780594609206835E11,5.823729435791173E11,5.866864262375511E11,5.90999908895985E11,5.953133915544188E11,5.996268742128527E11,6.039403568712865E11,6.082538395297203E11,6.125673221881542E11,6.16880804846588E11,6.211942875050219E11,6.255077701634557E11,6.298212528218895E11,6.341347354803234E11,6.384482181387572E11,6.42761700797191E11,6.470751834556249E11,6.513886661140587E11,6.557021487724926E11,6.600156314309264E11,6.643291140893602E11,6.686425967477941E11,6.729560794062279E11,6.772695620646617E11,6.815830447230956E11,6.858965273815294E11,6.902100100399633E11,6.94523492698397E11,6.988369753568308E11,7.031504580152646E11,7.074639406736985E11,7.117774233321323E11,7.160909059905662E11,7.20404388649E11,7.247178713074338E11,7.290313539658677E11,7.333448366243015E11,7.376583192827354E11,7.419718019411692E11,7.46285284599603E11,7.505987672580369E11,7.549122499164707E11,7.592257325749045E11,7.635392152333384E11,7.678526978917722E11,7.72166180550206E11,7.764796632086399E11,7.807931458670737E11,7.851066285255076E11,7.894201111839414E11,7.937335938423752E11,7.980470765008091E11,8.023605591592429E11,8.066740418176768E11,8.109875244761106E11,8.153010071345444E11,8.196144897929783E11,8.239279724514121E11,8.28241455109846E11,8.325549377682798E11,8.368684204267136E11,8.411819030851475E11,8.454953857435813E11,8.498088684020151E11,8.54122351060449E11,8.584358337188828E11,8.627493163773167E11,8.670627990357505E11,8.713762816941843E11,8.756897643526182E11,8.80003247011052E11,8.843167296694858E11,8.886302123279197E11,8.929436949863535E11,8.972571776447874E11,9.015706603032212E11,9.05884142961655E11,9.101976256200889E11,9.145111082785227E11,9.188245909369565E11,9.231380735953904E11,9.274515562538242E11,9.31765038912258E11,9.360785215706919E11,9.403920042291257E11,9.447054868875596E11,9.490189695459934E11,9.533324522044272E11,9.576459348628611E11,9.619594175212949E11,9.662729001797286E11,9.705863828381625E11,9.748998654965963E11,9.792133481550302E11,9.83526830813464E11,9.878403134718978E11,9.921537961303317E11,9.964672787887655E11,1.0007807614471993E12,1.0050942441056332E12,1.009407726764067E12,1.0137212094225009E12,1.0180346920809347E12,1.0223481747393685E12,1.0266616573978024E12,1.0309751400562362E12,1.03528862271467E12,1.0396021053731039E12,1.0439155880315377E12,1.0482290706899716E12,1.0525425533484054E12,1.0568560360068392E12,1.0611695186652731E12,1.0654830013237069E12,1.0697964839821407E12,1.0741099666405746E12,1.0784234492990084E12,1.0827369319574423E12,1.0870504146158761E12,1.0913638972743099E12,1.0956773799327438E12,1.0999908625911777E12,1.1043043452496116E12,1.1086178279080454E12,1.1129313105664792E12,1.117244793224913E12,1.121558275883347E12,1.1258717585417808E12,1.1301852412002146E12,1.1344987238586484E12,1.1388122065170823E12,1.143125689175516E12,1.14743917183395E12,1.1517526544923838E12,1.1560661371508176E12,1.1603796198092515E12,1.1646931024676853E12,1.1690065851261191E12,1.173320067784553E12,1.1776335504429868E12,1.1819470331014207E12,1.1862605157598545E12,1.1905739984182883E12,1.1948874810767222E12,1.199200963735156E12,1.2035144463935898E12,1.2078279290520237E12,1.2121414117104575E12,1.2164548943688914E12,1.2207683770273252E12,1.225081859685759E12,1.2293953423441929E12,1.2337088250026267E12,1.2380223076610605E12,1.2423357903194944E12,1.2466492729779282E12,1.250962755636362E12,1.255276238294796E12,1.2595897209532297E12,1.2639032036116636E12,1.2682166862700974E12,1.2725301689285313E12,1.276843651586965E12,1.281157134245399E12,1.2854706169038328E12,1.2897840995622666E12,1.2940975822207004E12,1.2984110648791343E12,1.302724547537568E12,1.307038030196002E12,1.3113515128544358E12,1.3156649955128696E12,1.3199784781713035E12,1.3242919608297373E12,1.3286054434881711E12,1.332918926146605E12,1.3372324088050388E12,1.3415458914634727E12,1.3458593741219065E12,1.3501728567803403E12,1.3544863394387742E12,1.358799822097208E12,1.3631133047556418E12,1.3674267874140757E12,1.3717402700725095E12,1.3760537527309434E12,1.3803672353893772E12,1.3846807180478108E12,1.3889942007062446E12,1.3933076833646785E12,1.3976211660231123E12,1.4019346486815461E12,1.40624813133998E12,1.4105616139984138E12,1.4148750966568477E12,1.4191885793152815E12,1.4235020619737153E12,1.4278155446321492E12,1.432129027290583E12,1.4364425099490168E12,1.4407559926074507E12,1.4450694752658845E12,1.4493829579243184E12,1.4536964405827522E12,1.458009923241186E12,1.4623234058996199E12,1.4666368885580537E12,1.4709503712164875E12,1.4752638538749214E12,1.4795773365333552E12,1.483890819191789E12,1.488204301850223E12,1.4925177845086567E12,1.4968312671670906E12,1.5011447498255244E12,1.5054582324839583E12,1.509771715142392E12,1.514085197800826E12,1.5183986804592598E12,1.5227121631176936E12,1.5270256457761274E12,1.5313391284345613E12,1.535652611092995E12,1.539966093751429E12,1.5442795764098628E12,1.5485930590682966E12,1.5529065417267305E12,1.5572200243851643E12,1.5615335070435981E12,1.565846989702032E12,1.5701604723604658E12,1.5744739550188997E12,1.5787874376773335E12,1.5831009203357673E12,1.5874144029942012E12,1.591727885652635E12,1.5960413683110688E12,1.6003548509695027E12,1.6046683336279365E12,1.6089818162863704E12,1.6132952989448042E12,1.617608781603238E12,1.6219222642616719E12,1.6262357469201057E12,1.6305492295785396E12,1.6348627122369734E12,1.6391761948954072E12,1.643489677553841E12,1.647803160212275E12,1.6521166428707087E12,1.6564301255291426E12,1.6607436081875764E12,1.6650570908460103E12,1.669370573504444E12,1.673684056162878E12,1.6779975388213118E12,1.6823110214797456E12,1.6866245041381794E12,1.6909379867966133E12,1.695251469455047E12,1.699564952113481E12,1.7038784347719148E12,1.7081919174303486E12,1.7125054000887825E12,1.7168188827472163E12,1.7211323654056501E12,1.725445848064084E12,1.7297593307225178E12,1.7340728133809517E12,1.7383862960393855E12,1.7426997786978193E12,1.7470132613562532E12,1.751326744014687E12,1.7556402266731208E12,1.7599537093315547E12,1.7642671919899885E12,1.7685806746484224E12,1.7728941573068562E12,1.77720763996529E12,1.7815211226237239E12,1.7858346052821577E12,1.7901480879405916E12,1.7944615705990254E12,1.7987750532574592E12,1.803088535915893E12,1.807402018574327E12,1.8117155012327607E12,1.8160289838911946E12,1.8203424665496284E12,1.8246559492080623E12,1.828969431866496E12,1.83328291452493E12,1.8375963971833638E12,1.8419098798417976E12,1.8462233625002314E12,1.8505368451586653E12,1.854850327817099E12,1.859163810475533E12,1.8634772931339668E12,1.8677907757924006E12,1.8721042584508345E12,1.8764177411092683E12,1.8807312237677021E12,1.885044706426136E12,1.8893581890845698E12,1.8936716717430037E12,1.8979851544014375E12,1.9022986370598713E12,1.9066121197183052E12,1.910925602376739E12,1.9152390850351729E12,1.9195525676936067E12,1.9238660503520405E12,1.9281795330104744E12,1.932493015668908E12,1.9368064983273418E12,1.9411199809857756E12,1.9454334636442095E12,1.9497469463026433E12,1.9540604289610771E12,1.958373911619511E12,1.9626873942779448E12,1.9670008769363787E12,1.9713143595948125E12,1.9756278422532463E12,1.9799413249116802E12,1.984254807570114E12,1.9885682902285479E12,1.9928817728869817E12,1.9971952555454155E12,2.0015087382038494E12,2.0058222208622832E12,2.010135703520717E12,2.014449186179151E12,2.0187626688375847E12,2.0230761514960186E12,2.0273896341544524E12,2.0317031168128862E12,2.03601659947132E12,2.040330082129754E12,2.0446435647881877E12,2.0489570474466216E12,2.0532705301050554E12,2.0575840127634893E12,2.061897495421923E12,2.066210978080357E12,2.0705244607387908E12,2.0748379433972246E12,2.0791514260556584E12,2.0834649087140923E12,2.0877783913725261E12,2.09209187403096E12,2.0964053566893938E12,2.1007188393478276E12,2.1050323220062615E12,2.1093458046646953E12,2.1136592873231292E12,2.117972769981563E12,2.1222862526399968E12,2.1265997352984307E12,2.1309132179568645E12,2.1352267006152983E12,2.1395401832737322E12,2.143853665932166E12,2.1481671485905999E12,2.1524806312490337E12,2.1567941139074675E12,2.1611075965659014E12,2.1654210792243352E12,2.169734561882769E12,2.174048044541203E12,2.1783615271996367E12,2.1826750098580706E12,2.1869884925165044E12,2.1913019751749382E12,2.195615457833372E12,2.1999289404918057E12,2.2042424231502397E12]\n",
       "}\n",
       "}],\n",
       "\"metainfo_list\":[],\n",
       "\"spec_id\":\"1\"\n",
       "};\n",
       "               window.letsPlotCall(function() {\n",
       "       \n",
       "               var toolbar = null;\n",
       "               var plotContainer = containerDiv;               \n",
       "               \n",
       "                   var options = {\n",
       "                       sizing: {\n",
       "                           width_mode: \"min\",\n",
       "                           height_mode: \"scaled\",\n",
       "                           width: width\n",
       "                       }\n",
       "                   };\n",
       "                   var fig = LetsPlot.buildPlotFromProcessedSpecs(plotSpec, -1, -1, plotContainer, options);\n",
       "                   if (toolbar) {\n",
       "                     toolbar.bind(fig);\n",
       "                   }\n",
       "               });\n",
       "               \n",
       "               break;\n",
       "           }\n",
       "       }\n",
       "   });\n",
       "   \n",
       "   observer.observe(containerDiv);\n",
       "   \n",
       "   // ----------\n",
       "   })();\n",
       "   \n",
       "   </script>"
      ],
      "text/plain": [
       "<lets_plot.plot.core.PlotSpec at 0x1f32dc8af90>"
      ]
     },
     "execution_count": 28,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "gapminder_52 = gapminder.query(\"year == 1952\")\n",
    "\n",
    "(\n",
    "  ggplot(gapminder_52, aes(\"gdp\")) +\n",
    "  geom_density() \n",
    ")"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "👥 **WORK IN PAIRS/GROUPS:**\n",
    "\n",
    "-   Add `scale_x_log10` to the plot.\n",
    "-   Specify `panel_grid_major_x=element_blank()` in the `theme` function.\n",
    "-   Edit the x and y axes using `labs`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Code here"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "DS202W",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.13.1"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
